/* changes RCA load records to Intel hex records */
/* Herb Johnson last updated Feb 28 2017 */
/* some code comes from c:\lcc\lib\wizard\textmode.tpl. */
/* also copied from unprn.c, assembler listing to .asm converter */
/* also copied from bin2hex.c (c) by bill beech, 2008 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <conio.h>

#define INBUF 32			/* number of input chars */
char strbuf[INBUF+5];		/* Console I/O buffer	*/

void Usage(char *programName)
{
	fprintf(stderr,"%s <infilename> usage:\n",programName);
	fprintf(stderr,"changes RCA hex records to Intel hex records.\n");
	fprintf(stderr,"outfilename is infilename+.hex.\n");
	fprintf(stderr,"asks before overwriting output file\n");
	fprintf(stderr,"%s -h prints this message.\n", programName);
	return;
}

/* returns the number of arguments up to two */
int HandleOptions(int argc,char *argv[])
{
	int i,options=0;

	for (i=1; i< argc;i++) {
		if (argv[i][0] == '/' || argv[i][0] == '-') {
			switch (argv[i][1]) {
				/* An argument -? means help is requested */
				case '?':
					Usage(argv[0]);
					return -1;
					break;
				case 'h':
				case 'H':
					if (!stricmp(argv[i]+1,"help")) {
						Usage(argv[0]);
						return -1;
						break;
					}

				default:
					fprintf(stderr,"unknown option %s\n",argv[i]);
					Usage(argv[0]);
					return -1;
					break;
			}
		}
		else { /* string is filename or number want exactly two */
			if (options==0 || options==1) options++;
			else {
				fprintf(stderr, "too many arguments\n");
				return -1;
			}
		}
	}
	if (options != 1) {
		fprintf(stderr, "too few or too many arguments\n");
		return -1;
	}
	return options; /* only returns 1 if successful */
}

char bin2hex(int value) {
	if ((value > 15) || (value < 0) ) {
		fprintf(stderr,"bad hex convert %i\n",value);
		return -1;
	}
	if ((value >=0) && (value <=9)) return ('0' + value);
	return (value-10 + 'A');
}

int printbuf(char buffer[], FILE *outfile) {
	/* process buffer to outfile line;  */
	/* input buffer is 80E0 B3 F8 F8 A3 E2 D3; */
	/* output outbuf  is :0680E000B3F8F8A3E2D39F */

	int buflen,cols;
	int goodcols, i, j;
	char thechar;
	int outerr;
	char bufout[512+80];
	int counter;
	int byte, nibble, address, chk, len;

	buflen = strlen(buffer);


	/* printf("strlen is %i, first chars are: ->%c%c%c<-\n", buflen, buffer[0], buffer[1], buffer[2]); */
	/* format "aaaa dd dd dd ... dd;" */

	if (buflen <= 4) { /* skip small strings*/
		fprintf(stderr,"skipped odd RCA record %s\n",buffer);
		return 0; /* output nothing */
	}

	nibble = address = byte = 0; 	counter = 0; chk = 0;
	len = -2; // minus 2 because we end up counting two address bytes!
	bufout[0] = ':'; bufout[1]=' '; /* empty INtel string */
	i = 3; // point to address
	for (j=0;j<buflen;j++) { /* process RCA record into binary and Intel too*/

		thechar = buffer[j];

		if ( ((thechar >= '0') && (thechar <= '9')) || ((thechar >= 'A') && (thechar <= 'F')) ) {
			/* hex chars only */
			if (thechar < 'A') nibble = thechar - '0';
			else nibble = thechar - 'A' + 10;
			byte = (byte*16) + nibble;

			if ( (j>=0) && (j<=3) ) { /* address */
				bufout[i++] = thechar;
				address = (address * 16) + nibble;
				counter++;
				if (j==3) i = 9; // point to data in Intel record
			}
			if (j>=4) { /* data by two's including addr */
				bufout[i++] = thechar;
				counter++;
			}
			if (counter >= 2) { /* byte value every two hex data or addr digits */
				chk -= byte & 0xff;
				len++;
				counter = 0; byte = 0;
			}
		}
		// spaces, ; etc. in RCA record are ignored.
	}
	/* end of input RCA string, will ignore semicolon */

	/* the checksum byte */
	/* i points to end of Intel string at checksum */
	chk -= len & 0xff; chk = chk & 0xFF;
	if (chk > 256)  {
		fprintf(stderr,"checksum too big, %i.\n", chk);
		return -1;
	}

	bufout[i]= bin2hex(chk/16); i++;
	bufout[i]= bin2hex(chk -((chk/16)*16)); i++;
	bufout[i] = '\n'; bufout[i+1] = '\0'; /* terminate output string */


	/* the length byte */
	if (len >= 256) {
		fprintf(stderr,"input record too long, %i hex bytes.\n", len);
		return -1;
	}
	bufout[1]= bin2hex(len/16);
	bufout[2]= bin2hex(len -((len/16)*16));

	/* type 00 record */
	bufout[7] = '0'; bufout[8] = '0';

	outerr = fprintf(outfile, bufout, "%s");
	if (outerr == EOF) {
		fprintf(stderr,"write error or EOF on output file\n");	return -1;
	}
	return 0;
}

int main(int argc,char *argv[])
{
	int option; /* index of options in command line */
	FILE *infile;
	FILE *outfile;
	char filename[80]; /* look for this file */
	char outname[85];
	int fnamelen; /* length of filename */
	char buffer[512+80];

	int bufind, buflen; /*index into buffer and length of string */
	int i, j, founddot;
	int startit; /* flag start of output */
	char Mcmd[5]; /* IDIOT command to start  */
	int namepos=0; /*location of filename in imagefile*/
	int cols, goodcols; /* number of left columns to delete, correct cols with tabs=8 */
	int overwrite; /* true 1 or false 0 to overwrite output file*/

	Mcmd[0] = '!'; Mcmd[1]='M'; Mcmd[2]= '\0'; /* IDIOT command to start  */

	if (argc == 1) {
		/* If no arguments we call the Usage routine and exit */
		Usage(argv[0]);
		return 1;
	}
	/* handle the program options */

	option = HandleOptions(argc,argv);
	/* returns -1 on error, 0 if help request or index to first filename*/
	if (option < 1) return -1; /* error */
	/*otherwise see HandleOption for syntax */
	/* printf("option 1 is filename %s\n", argv[1]); */

	/* process filename to find */
	fnamelen = strlen(argv[1]);
	if (fnamelen < 80) strcpy(filename, argv[1]); else {
		fprintf(stderr,"filename too long  %s\n", argv[1]);
		return -1;
	} /* else filetofindname is filename to look for AND output file */
	printf("input filename is *%s*\n", filename);

	/* create filename.hex for output, remove any other extension */
	founddot = 0;
	for (i=0;i<fnamelen;i++) {
		outname[i]=filename[i];
		if (outname[i]=='.') {
			founddot = 1;
			strcpy(outname+i+1, "hex\0");
			break;
		}
	}
	if (founddot==0) strcpy(outname+fnamelen, ".hex\n");
	printf("Output filename is *%s*\n",outname);


	/* open input  file for read */
	infile = fopen(filename,"r");
	if (infile == 0) {
		fprintf(stderr,"input file %s not found\n", filename);
		return -1;
	}

	/* printf("main: past openfile input file\n"); */


	/* open output file for read to see if exists */
	overwrite = 1;
	outfile = fopen(outname,"r");
	if (outfile != 0) { /* it exists */
		//fprintf(stderr,"output file %s exists, I won't overwrite.\n", outname);
		/* get yes or no input, set overwrite 1 for yes 0 for no */
		printf("Destination file %s exists. Overwrite? Y/N ", outname);
		strbuf[0]=INBUF; cgets(strbuf);
 		if ( (toupper(strbuf[2])) == 'N') return 0;
		overwrite = 1;

		/* close output file either way */
		if (fclose (outfile)!= 0) {
			fprintf(stderr,"output file opened for read %s not closed\n", outname);
			return -1;
		}
	} /* else output file does not exist, we will open it below */

	/* if no overwrite close files and exit*/
	if (overwrite == 0) {
		fprintf(stderr,"no overwrite, close and exit\n");
		fclose (infile);
		return -1;
	}

	/* if yes to overwrite, or new file, open for write and go on*/
	outfile = fopen(outname,"w");
	if (outfile == 0) {
		fprintf(stderr,"output file %s not found\n", outname);
		return -1;
		}

	/* printf("main: past openfile output file\n"); */

	/* read line by line, process, print line out */

	namepos = 0; buffer[0]=0; startit=1;
	while (1) {
		fgets(buffer, 512, infile);
		if (buffer == NULL) {
			fprintf(stderr,"input file read error\n");
			namepos = -1;
			break;
		}
		if (feof(infile)) {
			printf("\nEOF after %i lines\n",namepos);
			break;
		}
		if (startit) { //check first line
			if ((buffer[0] != '!') || (buffer[1] != 'M')) {
				fprintf(stderr,"input file doesn't start with !M\n");
				namepos = -1;
				break;
			}
			startit = 0;
		}
		if (buffer[0] == '!') { /* skip !M */
			if (printbuf(buffer+2,outfile) != 0) break;
		}
		else if (printbuf(buffer,outfile) != 0) break;

		namepos = namepos + 1;
		/* printf("."); /*show a trail*/
	}

	/* close files */
	if (fclose (infile)!=0) {
		fprintf(stderr,"input file %s not closed\n", argv[option]);
		return -1;
	}

	if (fclose (outfile)!=0) {
		fprintf(stderr,"output file %s not closed\n", outname);
		return -1;
	}
	/* printf("main: past closefile input & output file\n"); */
	return 0;
}
